import asyncio
import logging
import ssl
from typing import Dict, Optional

import aiohttp

from .behavior import Behavior
from .device import Device
from .execution import Execution
from .hue import Hue
from .hdmi import Hdmi
from .errors import raise_error, RequestError, Unauthorized
from .hsb_cacert import HSB_CACERT

MIN_API_LEVEL = 4

logger = logging.getLogger(__name__)


class HueSyncBox:
    """Control a Philips Hue Play HDMI Sync Box."""

    def __init__(
        self,
        host: str,
        id: str,
        access_token: Optional[str] = None,
        port: int = 443,
        path: str = "/api",
    ) -> None:
        self._host = host
        self._id = id
        self._access_token = access_token
        self._port = port
        self._path = path

        self._clientsession: aiohttp.ClientSession | None = None

        # API endpoints
        self.behavior: Behavior
        self.device: Device
        self.execution: Execution
        self.hdmi: Hdmi
        self.hue: Hue

        self._last_response = None  # For debugging purposes

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def _get_clientsession(self) -> aiohttp.ClientSession:
        """
        Get a clientsession that is tuned for communication with the Hue Syncbox
        """

        def _build_ssl_context() -> ssl.SSLContext:
            context = ssl.create_default_context(cadata=HSB_CACERT)
            context.hostname_checks_common_name = True
            return context

        # Creating an SSL context has some blocking IO so need to run it in the executor
        loop = asyncio.get_running_loop()
        context = await loop.run_in_executor(None, _build_ssl_context)

        connector = aiohttp.TCPConnector(
            enable_cleanup_closed=True,  # Home Assistant sets it so lets do it also
            ssl=context,
            limit_per_host=1,  # Syncbox can handle a limited amount of connections, only take what we need
        )

        return aiohttp.ClientSession(
            connector=connector, timeout=aiohttp.ClientTimeout(total=10)
        )

    @property
    def access_token(self) -> str | None:
        return self._access_token

    @property
    def last_response(self) -> Dict | None:
        return self._last_response

    async def is_registered(self):
        try:
            await self.request("get", "/registrations")
            return True
        except Unauthorized:
            return False
        return False

    async def register(
        self,
        application_name: str,
        instance_name: str,
        use_registered_token: bool = True,
    ):
        """
        Register with the huesyncbox

        application_name : Userfriendly name of your application
        instance_name : The specific instance of your application, e.g. a specific device the application is running on
        use_registered_token: When true use the token (if obtained) for subsequent requests

        returns registration info on success
        """
        response = await self.request(
            "post",
            "/registrations",
            {"appName": application_name, "instanceName": instance_name},
            auth=False,
        )  # Make sure to _not_ use a possibly invalid token as it will be rejected

        info = None
        if response:
            info = {
                "registration_id": response["registrationId"],
                "access_token": response["accessToken"],
            }

            if use_registered_token:
                self._access_token = info["access_token"]

        return info

    async def unregister(self, registration_id: str):
        """Unregister application from the huesyncbox, you can only unregister the id associated with the token in use."""
        await self.request("delete", f"/registrations/{registration_id}")

    async def initialize(self):
        await self.update()
        if self.device.api_level < MIN_API_LEVEL:
            logger.error(
                "This library requires at least API version %s. Please update the Philips Hue Play HDMI Sync Box.",
                MIN_API_LEVEL,
            )

    async def close(self):
        if self._clientsession is not None:
            await self._clientsession.close()

    async def update(self):
        response = await self.request("get", "")
        self._last_response = response

        if response:
            self.behavior = Behavior(response["behavior"], self.request)
            self.device = Device(response["device"], self.request)
            self.execution = Execution(response["execution"], self.request)
            self.hue = Hue(response["hue"], self.request)
            self.hdmi = Hdmi(response["hdmi"], self.request)

    async def request(
        self, method: str, path: str, data: Optional[Dict] = None, auth: bool = True
    ):
        """Make a request to the API."""

        if self._clientsession is None:
            self._clientsession = await self._get_clientsession()
            assert self._clientsession is not None

        if self._clientsession.closed:
            # Avoid runtime errors when connection is closed.
            # This solves an issue when Updates were scheduled and HA was shutdown
            return None

        url = f"https://{self._host}:{self._port}{self._path}/v1{path}"

        try:
            logger.debug("%s, %s, %s" % (method, url, data))

            headers = {"Content-Type": "application/json"}
            if auth and self._access_token:
                headers["Authorization"] = f"Bearer {self._access_token}"

            async with self._clientsession.request(
                method, url, json=data, headers=headers, server_hostname=self._id
            ) as resp:
                logger.debug("%s, %s" % (resp.status, await resp.text("utf-8")))

                data = None
                if resp.content_type == "application/json":
                    data = await resp.json()
                    if resp.status != 200:
                        if isinstance(data, dict):
                            _raise_on_error(data)
                        else:
                            logger.error(
                                "Received unexpected data format: %s" % str(data)
                            )
                return data
        except aiohttp.ClientError as err:
            logger.debug(err, exc_info=True)
            raise RequestError(f"Error requesting data from {self._host}") from err
        except asyncio.TimeoutError as err:
            logger.debug(err, exc_info=True)
            raise RequestError(f"Timeout requesting data from {self._host}") from err


def _raise_on_error(data: Dict):
    """Check response for error message."""
    raise_error(data["code"], data["message"])



from .helpers import generate_attribute_string


class Wifi:
    """Represent wifi status"""

    def __init__(self, raw) -> None:
        self._raw = raw

    @property
    def ssid(self) -> str:
        """Wifi SSID"""
        return self._raw["ssid"]

    @property
    def strength(self) -> int:
        """
        0 = not connected; 1 = weak; 2 = fair; 3 = good; 4 = excellent
        """
        return self._raw["strength"]


class Device:
    """Represent Device config."""

    def __init__(self, raw, request) -> None:
        self._raw = raw
        self._request = request
        self._wifi: Wifi | None = Wifi(self._raw["wifi"])

    def __str__(self) -> str:
        attributes = [
            "name",
            "device_type",
            "unique_id",
            "ip_address",
            "api_level",
            "firmware_version",
            "wifi",
        ]
        return generate_attribute_string(self, attributes)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Device):
            return NotImplemented
        return self._raw == other._raw

    @property
    def name(self) -> str:
        """Friendly name of the device."""
        return self._raw["name"]

    @property
    def device_type(self) -> str:
        """Device Type identifier."""
        return self._raw["deviceType"]

    @property
    def unique_id(self) -> str:
        """Capitalized hex string of the 6 byte / 12 characters device id without delimiters. Used as unique id on label, certificate common name, hostname etc."""
        return self._raw["uniqueId"]

    @property
    def ip_address(self) -> str:
        """Local IP address of the device."""
        return self._raw["ipAddress"]

    @property
    def api_level(self) -> int:
        """Supported API level of the device."""
        return self._raw["apiLevel"]

    @property
    def firmware_version(self) -> str:
        """User readable version of the device firmware, starting with decimal major .minor .maintenance format e.g. 1.12.3."""
        return self._raw["firmwareVersion"]

    @property
    def wifi(self) -> Wifi | None:
        """Root object for Wifi information if available."""
        return self._wifi

    @property
    def led_mode(self) -> int:
        """
        0 = off in powersave, passthrough or sync mode; 1 = regular; 2 = dimmed in powersave or passthrough mode and off in sync mode
        """
        return self._raw["ledMode"]

    async def set_led_mode(self, mode: int) -> None:
        await self._request("put", "/device", data={"ledMode": mode})

    async def update(self) -> None:
        response = await self._request("get", "/device")
        if response:
            self._raw = response
            self._wifi = Wifi(self._raw["wifi"]) if "wifi" in self._raw else None



from typing import Dict
from .helpers import generate_attribute_string


class Behavior:
    """Represent Behavior config of huesyncbox."""

    def __init__(self, raw, request) -> None:
        self._raw = raw
        self._request = request

    async def _put(self, data: Dict) -> None:
        await self._request("put", "/behavior", data=data)

    def __str__(self):
        attributes = [
            "forceDoviNative",
        ]
        return generate_attribute_string(self, attributes)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Behavior):
            return NotImplemented
        return self._raw == other._raw

    @property
    def force_dovi_native(self) -> int | None:
        """When the TV advertises Dolby Vision force to use native native mode. Disabled 0, Enabled 1."""
        return self._raw.get("forceDoviNative")

    async def set_force_dovi_native(self, enabled: int) -> None:
        """Force DolbyVision compatibility of huesyncbox on or off."""
        data = {"forceDoviNative": enabled}
        await self._put(data)

    async def update(self) -> None:
        response = await self._request("get", "/behavior")
        if response:
            self._raw = response



from .errors import AiohuesyncboxException as AiohuesyncboxException
from .errors import RequestError as RequestError
from .errors import Unauthorized as Unauthorized
from .errors import InvalidState as InvalidState

from .huesyncbox import HueSyncBox as HueSyncBox
from .hue import Group as Group
from .hue import Hue as Hue
from .behavior import Behavior as Behavior
from .device import Device as Device
from .device import Wifi as Wifi
from .execution import Execution as Execution
from .execution import SyncMode as SyncMode
from .hdmi import Hdmi as Hdmi
from .hdmi import Input as Input
from .hdmi import Output as Output

__all__ = [
    "AiohuesyncboxException",
    "RequestError",
    "Unauthorized",
    "InvalidState",
    "HueSyncBox",
    "Group",
    "Hue",
    "Behavior",
    "Device",
    "Wifi",
    "Execution",
    "SyncMode",
    "Hdmi",
    "Input",
    "Output",
]
